#include "AlarmTreeWidget.h"
#include "filter.h"
#include <QStringList>
#include<QHeaderView>
#include <QApplication>
#include <QDateTime>
#include <elettracolors.h>
#include "AlarmTreeWidgetItem.h"
#include <QMouseEvent>
#include <QMenu>
#include <QAction>
#include <QDesktopServices>
#include <QUrl>
#include <QScrollBar>
#include <QtDebug>

#include <cudata.h>
#include <cumacros.h>
#include <qustring.h>
#include <qustringlist.h>

using namespace std;

AlarmTreeWidget::AlarmTreeWidget(QWidget *parent) : QTreeWidget(parent)
{
    setColumnCount(Columns::COLUMN_COUNT);
    setColumnHidden(MICROSEC, true);
    setRootIsDecorated(false);
    setAutoScroll(true);
    setSelectionMode(QAbstractItemView::ExtendedSelection);
    setHeaderLabels(QStringList() << "Date/time" << "microsecs" << "Alarm" << "Status" << "Ack" << "Level" << "Sil. min." << "Device" << "Message");
    //    setSortingEnabled(true);
    header()->setSectionResizeMode(  QHeaderView::ResizeToContents);
    _filter_enabled = false;
    m_type = AlarmType;
    setMouseTracking(true);
    connect(this, SIGNAL(itemEntered(QTreeWidgetItem*,int)), this, SLOT(slotItemEntered(QTreeWidgetItem*,int)));
}


void AlarmTreeWidget::setType(AlarmTreeWidget::Type t) {
    m_type = t;
}

void AlarmTreeWidget::refresh(const CuData &v) {
    double d = static_cast<int>(v["timestamp_us"].toDouble());
    QString ts = QDateTime::fromSecsSinceEpoch(d).toString("HH:mm:ss");
    ts += QString("+%1").arg(d - static_cast<int>(d));
    // if a src in error is back online we must remove the item from the view
    QString src_fixed;
    QString src = QuString(v, "src");
    QString dev = src.section('/', 0, src.count('/') - 1);
    bool empty = QuString(v, "msg").contains(QRegularExpression("cannot\\s+extract"));
    bool err = v["err"].toBool() && !empty;
    QStringList data;
    setToolTip(QuString(v, "msg"));
    // error: err flag is true and message is not of type "cannot extract data..."
    // only AlarmType adds an entry for each error
    if (err && m_type == AlarmType) {
        if(!m_src_err_map.contains(src))
            m_src_err_map.insert(src, SrvErrNew|SrvErrUnack); // true: new
        data << m_gen_error_item(v);
    }
    else {
        if(m_src_err_map.contains(src)) {
            m_src_err_map.remove(src);
            src_fixed = src;
        }
        data = QuStringList(v, "value");
    }

    int scrollBarPos;
    QStringList fields;
    AlarmTreeWidgetItem* item = NULL;
    QStringList high, medium, low, lowest, unknown_level;
    // group by level in order to sort view
    foreach(QString row, data) {
        QString level = row.split("\t").at(AlarmTreeWidget::LEVEL);
        if(level.contains("high"))
            high << row;
        else if(level.contains("lowest"))
            lowest << row;
        else if(level.contains("medium"))
            medium << row;
        else if(level.contains("low"))
            low << row;
        else
            unknown_level << row;
    }
    data = unknown_level + lowest + low + medium + high;

    QString alanam, alkey; // alarm name, alarm key (=src/alarm_name)
    int index = 0;
    // sets of alarms for the given src
    QSet<QString> oldset = mSrcAlarmHash.value(src); // save current
    QSet<QString>& alsrcset = mSrcAlarmHash[src]; // modify with new data (see (*) )
    alsrcset.clear(); // empty set for src
    /* add items to the tree(s) if not already present */
    for (int i = data.size() - 1; i >= 0; i--) {
        fields = data[i].split("\t");
        alanam = fields.at(AlarmTreeWidgetItem::Alarm);
        alkey = m_make_key(alanam, src);
        /* save current scroll bar value */
        scrollBarPos = verticalScrollBar()->value();
        /* instead of GROUP information, show DEVICE */
        fields[DEVICE] = dev;
        item =  mItemHash.value(alkey);
        if(!item) /* create a new item */ {
            index = findPos(fields[0].toUInt(), fields[AlarmTreeWidgetItem::Level], fields[AlarmTreeWidgetItem::Status]);
            m_insert_item(src, fields, index);
        }
        else /* update fields of existing items */
            item->update(fields);

        /* restore scrollbar value */
        verticalScrollBar()->setValue(scrollBarPos);

        // (*) add alarm name to set of alarm names for this src
        alsrcset.insert(alanam); // alsrcset is a reference

        // notify new alarm detected. connected to Alarm::beep
        qDebug() << __PRETTY_FUNCTION__ << fields;
        if(fields.size() > 5 && (fields[3] == "ALARM" || fields[3] == "ERROR") && fields.last()  == "NEW")
            emit newAlarmDetected();
    }

    /* remove stale items: alarm names for src */
    QSet<QString> stale = oldset - alsrcset;
    m_remove_stale_items(stale, src); // works on mItemHash as well as on the tree
}


int AlarmTreeWidget::findPos(time_t ts, const QString &pri, const QString& status) {
    int i = 0, prio =  m_pri_to_int(pri), stat = m_status_to_int(status);
    //    printf("findPos called with ts %ld, pri %d\n", ts, prio);
    while(i < topLevelItemCount() && topLevelItem(i)->data(AlarmTreeWidgetItem::Status, Qt::UserRole).toInt() > stat) {
        //        printf("1st pri %d pri %d --> i %d\nok cont", prio, topLevelItem(i)->data(AlarmTreeWidgetItem::Level, Qt::UserRole).toInt(), i);
        i++;
    }

    while(i < topLevelItemCount() && topLevelItem(i)->data(AlarmTreeWidgetItem::Level, Qt::UserRole).toInt() > prio
          && topLevelItem(i)->data(AlarmTreeWidgetItem::Status, Qt::UserRole).toInt() == stat) {
        //        printf("1st pri %d pri %d --> i %d\n", prio, topLevelItem(i)->data(AlarmTreeWidgetItem::Level, Qt::UserRole).toInt(), i);
        i++;
    }
    // i points to the first item with same pri or to the very last item in the view

    while(i < topLevelItemCount() && topLevelItem(i)->data(AlarmTreeWidgetItem::DateTime, Qt::UserRole).toUInt() > ts
          && topLevelItem(i)->data(AlarmTreeWidgetItem::Level, Qt::UserRole).toInt() == prio
          && topLevelItem(i)->data(AlarmTreeWidgetItem::Status, Qt::UserRole).toInt() == stat) {
        //        printf("1st ts %u item ts %u --> i %d\n", ts, topLevelItem(i)->data(AlarmTreeWidgetItem::DateTime, Qt::UserRole).toUInt(), i);
        i++;
    }
    return i;
}

int AlarmTreeWidget::countAll() const {
    return topLevelItemCount();
}

int AlarmTreeWidget::countAlarmed() const {
    int cnt = 0;
    for(int i = 0; i < topLevelItemCount(); i++) {
        const QString& s = topLevelItem(i)->text(AlarmTreeWidgetItem::Status);
        if( (s == "ALARM" || s == "ERROR") && topLevelItem(i)->text(AlarmTreeWidgetItem::Ack) == "NACK")
            cnt++;
    }
    return cnt;
}

void AlarmTreeWidget::setServerErrNotNew(const QString &errdev) {
    if(m_src_err_map.contains(errdev))
        m_src_err_map[errdev] &= ~SrvErrNew;
}

void AlarmTreeWidget::setServerErrAck(const QString &n) {
    if(m_src_err_map.contains(n))
        m_src_err_map[n] &= ~SrvErrUnack;
}

void AlarmTreeWidget::contextMenuEvent(QContextMenuEvent *e)
{
    bool silenceAvail = true;
    bool enable = true;
    QList<QTreeWidgetItem *>items = this->selectedItems();
    foreach(QTreeWidgetItem *it, items) {
        int sil = it->text(AlarmTreeWidgetItem::SilencedMin).toInt();
        silenceAvail &= (sil == 0);
        enable &= (sil > 0);
    }
    if(items.size()) {
        QMenu *menu = new QMenu(this);
        menu->addAction("Acknowledge", this, SIGNAL(ack()));
        if(silenceAvail)
            menu->addAction("Silence", this, SIGNAL(silence()));

        QTreeWidgetItem* it = itemAt(e->pos());
        if(it && it->text(MESSAGE).contains("   ")) {
            QAction* wikiAction = menu->addAction("Wiki...");
            wikiAction->setProperty("url", it->text(MESSAGE).split("   ").last());
            connect(wikiAction, SIGNAL(triggered()), this, SLOT(openUrl()));
        }

        // enable
        if(m_type == AlarmType && enable)
            menu->addAction("Enable", this, SIGNAL(enable()));

        menu->exec(QCursor::pos());
        delete menu;
    }
}

void AlarmTreeWidget::m_insert_item(const QString& src, const QStringList &list, int index)
{
    AlarmTreeWidgetItem *item;
    QStringList fields(list);
    bool isNew = false;
    if(fields.size() > 0)
        fields[AlarmTreeWidgetItem::DateTime] = QDateTime::fromTime_t(list.at(AlarmTreeWidgetItem::DateTime).toUInt()).toString();

    if(fields.at(AlarmTreeWidgetItem::Alarm).contains("internal_error_"))
        item = new AlarmTreeWidgetItem(this, fields);
    else
    {
        if(fields.last().toUpper() == "NEW"){
            isNew = true;
        }
        fields.removeLast(); /* the last element contains "NEW" flag, need not be displayed */
        item = new AlarmTreeWidgetItem(fields);
        item->setData(AlarmTreeWidgetItem::Level, Qt::UserRole, m_pri_to_int(list[AlarmTreeWidgetItem::Level]));
        item->setData(AlarmTreeWidgetItem::DateTime, Qt::UserRole, list.at(AlarmTreeWidgetItem::DateTime).toUInt());
        item->setData(AlarmTreeWidgetItem::Status, Qt::UserRole, m_status_to_int(list.at(AlarmTreeWidgetItem::Status)));
        QString s = src.section('/', 0, src.count('/') - 1);
        QRegularExpression re("(?:tango://)*(.*):\\d+/(.*)");
        QRegularExpressionMatch ma = re.match(s);
        if(ma.capturedTexts().size() > 2) {
            QString s = ma.capturedTexts().at(1);
            s.contains('.') ? s = "[" + s.section('.', 0, 0) + "] " + ma.capturedTexts()[2] : s = s;
        }
        item->setData(AlarmTreeWidgetItem::Device, Qt::UserRole, src.section('/', 0, src.count('/') - 1));
        item->setText(AlarmTreeWidgetItem::Device, s);
        insertTopLevelItem(index, item);
        item->setNew(isNew);
    }
    /* add the couple identifier/AlarmTreeWidgetItem to the hash */
    mItemHash.insert(m_make_key(fields.at(AlarmTreeWidgetItem::Alarm), src), item);
}

QString AlarmTreeWidget::m_gen_error_item(const CuData &da) const {
    QStringList l;
    QDateTime dt = QDateTime::currentDateTime();
    QString dev, src = QuString(da, "src"), msg = QuString(da, "msg");
    dev = src.section('/', 0, dev.count('/') - 1);
    qDebug() << __PRETTY_FUNCTION__ << m_src_err_map;
    bool _new = ((m_src_err_map[src] & (SrvErrNew | SrvErrUnack)) == (SrvErrNew | SrvErrUnack));
    printf("\e[1;36mAlarmTreeWidget::m_gen_error_item %d  & %d\e[0m\n", m_src_err_map[src] , (SrvErrNew | SrvErrUnack));
    l << QString::number(dt.toSecsSinceEpoch()) << "-" << src << "ERROR" <<  (m_src_err_map[src] & SrvErrUnack ? "UNACK" : "ACK") << "high" << "-1" << "gr_handler"
        << "Alarm server error: " + msg.replace("\n", " ") << (_new ? "NEW" : " ");
    qDebug() << __PRETTY_FUNCTION__ << l;
    return l.join("\t");
}

AlarmTreeWidgetItem *AlarmTreeWidget::m_findItem(const QString &alarm, const QString &src) const {
    return mItemHash.value(m_make_key(alarm, src));
}

QString AlarmTreeWidget::m_make_key(const QString &alarm, const QString &src) const {
    return QString("%1/%2").arg(src).arg(alarm);
}


void AlarmTreeWidget::m_remove_stale_items(const QSet<QString> stale_names, const QString& src) {
    foreach(const QString& name, stale_names) {
        const QString& key = m_make_key(name, src);
        AlarmTreeWidgetItem *item = mItemHash.value(key);
        if(item) {
            mItemHash.remove(key);
            delete item;
        }
    }
}

int AlarmTreeWidget::m_pri_to_int(const QString &pri) const {
    if(pri == "highest") return 150;
    else if(pri == "high") return 100;
    else if(pri == "medium") return 80;
    else if(pri == "low") return 50;
    else return 10;
}

int AlarmTreeWidget::m_status_to_int(const QString &stat) {
    if(stat == "ALARM")
        return 100;
    else if(stat == "ERROR")
        return 80;
    return 50;
}

void AlarmTreeWidget::mouseMoveEvent(QMouseEvent *event)
{
    QTreeWidget::mouseMoveEvent(event);
    QPoint p = event->pos();
    emit mouseOverItem(this->itemAt(p) != NULL);
}

void AlarmTreeWidget::slotItemEntered(QTreeWidgetItem* it, int)
{
    QString mes = it->text(AlarmTreeWidgetItem::Message);
    emit message(mes);
}

void AlarmTreeWidget::openUrl()
{
    QObject *send = sender();
    if(send->property("url").isValid()) {
        QDesktopServices::openUrl(QUrl(send->property("url").toString()));
    }
}


